from ruis import DirType, FileType
import socket
import thread
import time

# unescaping: %20 -> space etc.
from urllib import unquote as _unescape

# Some thoughts on further development
#
# - tta: textbox contents, attributes etc. should not be sent from the
# browser to the server on every event. instead, the server could
# perhaps ask for them if it wants them like
# add_to_browser_queue('send_value(itemname);')

class WebformpError(Exception): pass

_browser_side_js = file("webformp.js").read()

_html_basepage='''
<html><body>

<form name="form0">
%s
</form>

<script language="JavaScript" type="text/javascript">
%s
</script>

</body></html>
'''

default_favicon = '\x89PNG\r\n\x1a\n\x00\x00\x00\rIHDR\x00\x00\x00\x10\x00\x00\x00\x10\x01\x03\x00\x00\x00%=m"\x00\x00\x00\x06PLTE\x00\x03\x00\x00\x00\x00\xe2\xc7\xc3\x1f\x00\x00\x00\x01tRNS\x00@\xe6\xd8f\x00\x00\x00\x01bKGD\x00\x88\x05\x1dH\x00\x00\x00\tpHYs\x00\x00\x0b\x13\x00\x00\x0b\x13\x01\x00\x9a\x9c\x18\x00\x00\x005IDAT\x08\xd7c8\xd0\xc0P\xd0\xc0\xd0\xdf\xc0\xf0\xf0\x01\x83\xdc\x1c\x86\xc3\xc7\x19lv0\xf4\xf73\x14|\x00q\x1f>\x06q\x99\x8f30|``\xe8g`h\x06\x00\xb83\x10\xf5\xe3`$\'\x00\x00\x00\x00IEND\xaeB`\x82'

def _http_send_ok(conn, s):
    resp = "HTTP/1.1 200 OK\r\nServer: Ruis\r\nKeep-Alive: timeout=30\r\nConnection: Keep-Alive\r\nContent-Type: text/html; charset=iso-8859-1\r\nContent-Length: %s\r\n\r\n%s" % (len(s),s)
    return conn.sendall(resp)

def _http_send_404(conn, s):
    resp = "HTTP/1.1 404 Not found\r\nServer: Ruis\r\nKeep-Alive: timeout=30\r\nConnection: Keep-Alive\r\nTransfer-Encoding: chunked\r\nContent-Type: text/html; charset=iso-8859-1\r\n\r\n%x\r\n%s0\r\n\r\n" % (len(s),s)
    return conn.sendall(resp)

def _http_read(conn):
    # FIX ME! read and return proper packets! Educated guess, but
    # stupid assumption: header is received completely on the first
    # recv, data may have to be recv'd again. Can do content-length
    # --- only.
    data = conn.recv(2048)
    while data != "" and not (data.startswith("GET") or data.startswith("POST")):
        data = conn.recv(2048)
    if data == "": return ""
    if data.startswith("POST"): # require content-length in POST
        while not "content-length" in data.lower():
            data += conn.recv(2048)
    datarows = data.split('\n')
    datalength = 0
    datahead, databody = data, ""
    for row in datarows:
        if row.lower().startswith('content-length'):
            datalength = int(row.strip().split()[1])
            datahead, databody = data.split('\r\n\r\n',1)
    while len(databody) < datalength:
        databody += conn.recv(2048)

    # FIX: favicon.ico should be an optional parameter in webformp.create
    if "favicon.ico" in datahead:
        _http_send_ok(conn, default_favicon)
        return _http_read(conn)
    return datahead + '\r\n\r\n' + databody

class webformp(DirType):
    """
    Web Form Protocol
    
    Creating a web form connection gives an empty web page to the
    connected browser and a web form connection (webforms)
    subdirectory under webformp directory. Webforms can be created
    under the webforms directory. Create-method there is given the
    form contents as an argument.
    """

    def __init__(self, *a, **kw):
        DirType.__init__(self, *a, **kw)
        self.name="webformp"
        self._sockets = {}
        self._sid2wfs = {} # map session-id's to webforms(ession) objects
        self._allow_new_session_lock = thread.allocate_lock()
        self._allow_new_session = {} # port -> [lock1, lock2, ...]

    def create(self, connspec, **kw):
        """
        /conn/webformp.create(connspec)
        Example of connspec: 'listen:tcp:5555'
        """
        if connspec[:11] == "listen:tcp:":
            try:
                host, sport = connspec[10:].split(":")
                port = int(sport)
            except Exception, e:
                raise Exception("Error in connspec: %s" % e)
            new_session_lock = self._new_session(port)
            new_session_lock.lock.acquire()
            if new_session_lock.wfsname == "":
                raise WebformpError("Cannot accept TCP connections from spec: %s" % (connspec,))
            else:
                return new_session_lock.wfsname
        else:
            return None

    def _new_session(self, port):
        """returns a lock that is acquired when the new session
        has been established"""
        class sessionlock(object): pass
        sl = sessionlock()
        sl.lock = thread.allocate_lock()
        sl.lock.acquire()
        sl.wfsname = ""
        self._allow_new_session_lock.acquire()
        if not port in self._allow_new_session:
            # There is no server thread listening to the port.
            # Let's start one.
            self._allow_new_session[port] = [sl]
            thread.start_new_thread(self._http_server, (port,))
        else:
            # Allow existing server thread, which listens to the given
            # port, accept a new session.
            self._allow_new_session[port].append(sl)
        self._allow_new_session_lock.release()
        return sl
        
    def _new_session_allowed(self, port):
        """returns 1 if new session through the port is allowed,
        otherwise 0"""
        self._allow_new_session_lock.acquire()
        if len(self._allow_new_session[port]) > 0:
            first_lock = self._allow_new_session[port].pop(0)
            self._allow_new_session_lock.release()
            return first_lock
        else:
            self._allow_new_session_lock.release()
            return None

    def _http_server(self, port):
        # There is one http_server thread for each port that is
        # listened. These threads start new threads for handling new
        # tcp connections.
        while 1:
            try:
                conn = self._accept_tcp_connection(port)
            except:
                for l in self._allow_new_session[port]:
                    l.lock.release()
                break

            # given a tcp connection, start a new server thread for
            # that connection in the connection threads, keep track on
            # the number of connections per session session id is
            # carried on every url and it is generated by this server
            # for new sessions

            # URL types:
            # New session: GET /
            # Event:       GET /session-id/object.method
            # Form data:   POST /session-id/FORM_ELEMENTS (post data defines the elements)
            # Erroneous:   all other URLs

            # session-id starts with "session-".

            thread.start_new_thread(self._serve_connection, (conn, port))
            
    def _serve_connection(self, conn, port):
        # Connection handler. This handles http messages received
        # from the connection, establishes new web form sessions and
        # routes http messages to the correct sessions.
        while 1:
            data = _http_read(conn)
            if data == "": # connection lost
                conn.close()
                break
            
            session, object_id = self._check_session(data)
            if not session: # this starts a new session
                session_lock = self._new_session_allowed(port)
                if not session_lock:
                    # print "new session not allowed to", port, data.split('\n')[:4]
                    _http_send_404(conn, "")
                    conn.close()
                    break
                else: # register new session and release the lock
                    wfs = webforms()
                    self._register_to_dir(str(wfs), wfs)
                    # THINK: need for cryptic session id?
                    identified_session = "session-%s" % (str(wfs),)
                    wfs._set_session_id(identified_session)
                    self._sid2wfs[identified_session] = wfs
                    response = _html_basepage % ("", _browser_side_js %
                                                 (identified_session,))
                    session_lock.wfsname = str(wfs)
                    session_lock.lock.release()
            elif session in self._sid2wfs: # now the session is identified
                # check connection validity in response because a
                # browser can use the same tcp connection for
                # different sessions.
                #response = ('if (session_id == "%s") {' % (session+"."+form,)) + \
                #           self._sid2wfs[session]._handle_http_data(data) + \
                #           '}'
                response = self._sid2wfs[session]._handle_http_data(data)
            else: # session cannot be found in session library. strange
                print "UNEXPECTED SESSION:", session, data.split('\n')[0]
                response = ""
            _http_send_ok(conn, response)

    def _check_session(self, s):
        """returns id of an session (a non-empty string) or
        an empty string if data seems to start a new session"""
        # it's the string between the first and the second slash,
        # either limited by the second slash or a dot that separates
        # session id from form id.
        s = s.split('\n',1)[0] # it's on the first row
        s = s.split(' ')[1] # between two spaces: GET<space>addr<space>HTTP/1.1
        session_id = ""
        object_id = ""
        try: 
            first_slash = s.index("/")
            second_slash = s.index("/", first_slash+1)
            session_id = s[first_slash+1:second_slash]
            object_id = s[second_slash+1:s.index(" ", second_slash+1)]
        except ValueError: pass
        return session_id, object_id

    def _accept_tcp_connection(self, port):
        """
        Returns an open socket object.
        """
        if port in self._sockets:
            s = self._sockets[port]
        else:
            s = socket.socket(socket.AF_INET,socket.SOCK_STREAM)
            s.bind(('',port))
            s.listen(1)
            self._sockets[port] = s
        conn, src = s.accept()
        # print "CONNECTION FROM",src
        return conn
    

class webforms(DirType):
    """
    Web Form Session handles http requests directed to this session.
    Forms can be created under this directory with create(htmlform)
    method. Active form can be queried with get_active() method and
    changed with set_active(formfilename) method.
    """
    counter = 0
    
    def __init__(self, **kw):
        DirType.__init__(self, **kw)

        self.name = str(webforms.counter) # FIX RACE on counter
        webforms.counter += 1

        self._fso2eltid = {} # map file system object to element id in www form
        self._eltid2fso = {}
        self._to_browser_queue = []
        self._to_browser_queue_lock = thread.allocate_lock()
        self._active_webform = webform(self) # a dummy form
        self._form_changing_to = None # str(self._active_webform)
        self._activating_new_form_lock = thread.allocate_lock()
        self._activating_new_form_lock.acquire()

    def _set_session_id(self, session_id):
        self._session_id = session_id

    def create(self, htmlform):
        newform = webform(self)
        newform._html = htmlform
        self._register_to_dir(str(newform), newform)
        self.set_active(str(newform))
        return str(newform)

    def set_active(self, form_id):
        """
        Activates form_id. The browser should be evaluating (eval)
        responses.
        """
        if not form_id in self._contents:
            raise Exception("Form %s not found in %s" %
                            (form_id, ', '.join(self.fullname)))
        self._active_webform = self._contents[form_id]

        self._add_to_browser_queue(
            "document.forms[0].innerHTML='%s'; bind_events();" %
            (self._active_webform._html.replace('\n',' ').replace('\r',''),))

        if self._active_webform._elements_received == 0:
            # does not return until new form elements have been parsed
            # because the form directory does not have any contents
            # before that
            self._add_to_browser_queue("send_form_elements();")
            self._activating_new_form_lock.acquire()
        
    def _response_from_browser_queue(self):
        self._to_browser_queue_lock.acquire()
        response = ';\n\n'.join(self._to_browser_queue)
        self._to_browser_queue = []
        self._to_browser_queue_lock.release()
        return response

    def _add_to_browser_queue(self, js):
        self._to_browser_queue_lock.acquire()
        try:
            self._to_browser_queue.append(js)
        finally: self._to_browser_queue_lock.release()

    def _handle_http_data(self, data):
        
        def check_form_elements(data):
            # if string FORM_ELEMENTS appears in the first row,
            # form data is coming up
            if "FORM_ELEMENTS" in data.split('\n',1)[0]:
                self._active_webform._parse_formdata(data)
                return 1
            return 0

        def check_event(data):
            """
            If something must be responded to the browser right away,
            it is returned as a return value. This should be done only
            when the browser requests the whole page. Those responses
            are in HTML. Normal responses (Javascript) is sent by
            putting the data in 'to_browser_queue'.
            """
            # parse event
            try:
                # FIX: repeating code (check_session) --- implement
                # URL parsing in one place only.
                first_slash = data.index("/")
                second_slash = data.index("/", first_slash+1)
                dot_after_slashes = data.index(".", second_slash)
                space_after_slashes = data.index(" ", second_slash)
                session_id = data[first_slash+1:second_slash]
                if space_after_slashes < dot_after_slashes:
                    objname = ""
                    methodname = ""
                    methodarg = ""
                else:
                    objname = data[second_slash+1:dot_after_slashes]
                    methodname = data[dot_after_slashes+1:space_after_slashes]
                    methodarg = data[data.index('\r\n\r\n')+4:].strip()
            except Exception, e:
                print "Cannot parse event from data", e, data

            if objname == "":
                # matching form_id but request has no
                # parameters. Assume that reload has been pressed.
                # Response must be urgent, because the browser expects
                # to get all the page at once.
                thread.start_new_thread(self._active_webform.browser_reload, ())
                # TODO: send the base page and the html of the current form
                # *in the same package* to save time
                return None
                
            # try to send the event to self._active_webform:
            elif objname in self._active_webform._contents:
                try:
                    # print '%s.%s("%s")' % (objname, methodname, methodarg)
                    method = getattr(self._active_webform._contents[objname],
                                     methodname)
                    thread.start_new_thread(method, (methodarg,),
                                            {'synthetic':0})
                    return 1
                except Exception,e:
                    print e
                    pass
            return None

        # The following lines should be executed reasonably
        # quickly in order to send response in time and thereby
        # avoid browser queueing ajax requests. Everything that
        # may take time should be done in a separate thread.

        # Does this message contain form elements?
        content_identified = check_form_elements(data)
        
        # If not, check if this message is an event
        if not content_identified:
            content_identified = check_event(data)
            if type(content_identified)==str: # there is an urgent message
                return content_identified

        if not content_identified:
            # print "http_server: could not handle message"
            pass

        return self._response_from_browser_queue()


class webform(DirType):

    counter = -1 # the first instance will be "dummy", not worth of a natural number

    def __init__(self, wfs_obj, **kw):
        """takes html and webforms object (parent) as arguments"""
        DirType.__init__(self, **kw)

        self._wfs_obj = wfs_obj
        self._elements_received = 0
        self.name = str(webform.counter)
        webform.counter += 1
    
    def _handle_event(self, elt_type=None, elt_name=None, elt_id=None, elt_event=None, elt_value=None, synthetic=1):
        """
        this is called when an event is received from the web form (in
        this case, keyword argument 'synthetic' is set to 0), or when
        an event is caused by calling a method through the form
        element object in ruis fs.
        """
        if synthetic==1: # this information should be passed to the web form
            if elt_value != None: # value should be passed to the form
                # escape because it's going to be java script
                # how about JSON to do this correctly?
                elt_value = elt_value.replace('\n','\\n')
                self._wfs_obj._add_to_browser_queue(
                    'document.getElementById("%s").value="%s";' %
                    (elt_id, elt_value))
        else:
            pass # callbacks may do what they do, this guy does nothing.

    def _parse_formdata(self, formdata):
        # CONSIDER: json might be much smarter way to transfer this data
        elt_id, elt_type, elt_name, elt_value, elt_events = [""]*5
        for row in formdata.split('\n'):
            row = row.strip()
            if row.startswith('ID '): elt_id = row[3:]
            elif row.startswith('TYPE '): elt_type = row[5:]
            elif row.startswith('NAME '): elt_name = row[5:]
            elif row.startswith('VALUE '): elt_value = row[6:]
            elif row.startswith('EVENTS '): elt_events = row[7:].split()
            elif row.startswith('COMMIT'):
                # Create a new class for each form item. Add methods
                # that correspond to registered events.
                class webform_item(webform_itembase): pass
                obj = webform_item(elt_id, elt_type, elt_name, elt_value)
                for elt_event in elt_events:
                    def generic_method(_self, elt_value=None, synthetic=1):
                        if elt_value != None:
                            _self.elt_value = _unescape(elt_value)
                        return self._handle_event(_self.elt_type, _self.elt_name, _self.elt_id, elt_event, elt_value, synthetic)
                    generic_method.func_name = elt_event
                    setattr(webform_item, elt_event, generic_method)
                self.create(obj)
                self._wfs_obj._fso2eltid[str(obj)] = elt_id
                self._wfs_obj._eltid2fso[elt_id] = obj
                elt_id, elt_type, elt_name, elt_value, elt_events = [""]*5
        if self._elements_received == 0:
            # this is the first time receiving elements, so
            # webforms.set_active must be waiting for this...
            self._wfs_obj._activating_new_form_lock.release() # let set_active continue
            self._elements_received = 1

    def browser_reload(self):
        """Reload-button pressed"""
        pass

class webform_itembase(FileType):
    def __init__(self, elt_id, elt_type, elt_name, elt_value):
        FileType.__init__(self, name=elt_name)
        self.elt_id = elt_id
        self.elt_type = elt_type
        self.elt_name = elt_name
        self.elt_value = elt_value

    def __str__(self):
        return self.elt_name

    def get_value(self):
        return self.elt_value
